#include <stdio.h>
#include <errno.h>
#include "DEFS.h"
#include "PATH.h"

PATH *CreatePath(int, int, int);
void PATH_Free(PATH *);
int GetCurrentTimeSlice(PATH *, REAL *);
int GetCurrentMCStep(PATH *, REAL *);
int GetPreviousTimeSlice(PATH *, REAL *);
int GetNextTimeSlice(PATH *, REAL *);
void FlipAllHistories(PATH *);
void UpdateCurrentTimeSlice(PATH *, REAL *);
void ShiftCurrentRateHistory(PATH *, REAL );
void ShiftCurrentRate(PATH *, REAL );
int SetCurrentPath(PATH *, REAL * );
void GetCurrentPath(PATH *, REAL * );
void GetInitLagsofCurrentPath(PATH *, REAL * );
void SetInitLagsofCurrentPath(PATH *, REAL * );
int SetStraightPath(PATH *, REAL );
void SetAllPathsToZero(PATH *);
int SetCurrentTimeSlice(PATH *, REAL * );
int SetInitLagsToZero(PATH *);
int SetAllInitLags(PATH *, REAL ** );
int SetInitLag(int, PATH *, REAL *);
void SetTimeGridSize(PATH *, int);
void SetRateGridSize(PATH *, int);
void FetchRunningTimeSlice(PATH *, REAL * );
void PATH_Add(PATH *, PATH *);



#ifdef INLINE

#define ProcessOrder(PathParams) ((PathParams)->Lag)
#define HowManyRates(PathParams) ((PathParams)->RateGridSize)
#define HowManyTimeSteps(PathParams) ((PathParams)->TimeGridSize)
#define SetCurrentTimetoInitial(PathParams) ((PathParams)->CurrentTimeIndex = 0)
#define SetCurrentTimeIndex(PathParams,j)   if(((j)>(PathParams)->TimeGridSize) || ((j)<0)) exit(1);  (PathParams)->CurrentTimeIndex = (j);
#define GetCurrentTimeIndex(PathParams) ((PathParams)->CurrentTimeIndex)
#define SetRunningTimetoCurrent(PathParams) (((PathParams)->RunningTimeIndex) = GetCurrentTimeIndex(PathParams))
#define SetRunningTimeOneStepBack(PathParams) (((PathParams)->RunningTimeIndex) = (((PathParams)->RunningTimeIndex)-1))
#define GetPreviousTimeIndex(PathParams) (((PathParams)->CurrentTimeIndex) - 1)
#define GetNextTimeIndex(PathParams) (((PathParams)->CurrentTimeIndex) + 1)
#define SetCurrentTimetoNext(PathParams) ((PathParams)->CurrentTimeIndex = (PathParams)->CurrentTimeIndex + 1)
#define BeyondHorizon(PathParams) ((GetCurrentTimeIndex(PathParams)) > (HowManyTimeSteps(PathParams)) ? 1: 0)
#define OnHorizon(PathParams) ((GetCurrentTimeIndex(PathParams)) == (HowManyTimeSteps(PathParams)) ? 1: 0)
#define SetCurrentRatetoFirst(PathParams) ((PathParams)->CurrentRate = 1)
#define SetCurrentRateIndex(PathParams, ir) (((PathParams)->CurrentRate) = (ir))
#define GetCurrentRateIndex(PathParams) ((PathParams)->CurrentRate)
#define FinishedLastRate(PathParams) (((PathParams)->CurrentRate) > ((PathParams)->RateGridSize) ? 1 : 0)
#define SetCurrentRatetoNext(PathParams) (((PathParams)->CurrentRate) = ((PathParams)->CurrentRate + 1))

#else

int ProcessOrder( PATH *);
int HowManyRates( PATH *);
int HowManyTimeSteps( PATH *);
void SetCurrentTimetoInitial( PATH *);
void SetCurrentTimeIndex( PATH *, int);
int GetCurrentTimeIndex(PATH *);
void SetRunningTimetoCurrent(PATH *);
void SetRunningTimeOneStepBack(PATH *);
int GetPreviousTimeIndex(PATH *);
int GetNextTimeIndex(PATH *);
void SetCurrentTimetoNext(PATH *);
int BeyondHorizon(PATH *);
int OnHorizon( PATH *);
void SetCurrentRatetoFirst( PATH *);
void SetCurrentRateIndex(PATH *, int );
int FinishedLastRate( PATH *);
int GetCurrentRateIndex( PATH *);
void SetCurrentRatetoNext( PATH *);



int ProcessOrder( PATH *PathParams)
{
  return PathParams->Lag;
}



int HowManyRates( PATH *PathParams)
{
  return PathParams->RateGridSize;
}




int HowManyTimeSteps( PATH *PathParams )
{
  return PathParams->TimeGridSize;
}




void SetCurrentTimetoInitial( PATH *PathParams )
{
  PathParams->CurrentTimeIndex = 0;
}


void SetCurrentTimeIndex
                       ( 
                       PATH *PathParams,
                       int j
                       )
{
  if((j>PathParams->TimeGridSize) || (j<0)) exit(1);
  PathParams->CurrentTimeIndex = j;
}


int GetCurrentTimeIndex(PATH *PathParams )
{
  return PathParams->CurrentTimeIndex;
}




void SetRunningTimetoCurrent(PATH *PathParams)
{
 int itime=GetCurrentTimeIndex(PathParams);
 PathParams->RunningTimeIndex = itime;
}




void SetRunningTimeOneStepBack(PATH *PathParams)
{
 int itime = PathParams->RunningTimeIndex;
 PathParams->RunningTimeIndex = --itime;
}




int GetPreviousTimeIndex(PATH *PathParams )
{
  int itp = PathParams->CurrentTimeIndex;
  if(--itp < 0) {
   fprintf(stderr,"GetPreviousTimeIndex(): Gone beyond initial time slice\n");
   exit(1);
  }
  return itp;
}



int GetNextTimeIndex(PATH *PathParams )
{
  int itp = PathParams->CurrentTimeIndex;
  if(++itp > (HowManyTimeSteps(PathParams) + 1)) {
   fprintf(stderr,"GetNextTimeIndex(): Alredy gone beyond horizon \n");
   exit(1);
  }
  return itp;
}



void SetCurrentTimetoNext(PATH *PathParams )
{
  int it = GetCurrentTimeIndex(PathParams);
  int NumTimeSteps=HowManyTimeSteps(PathParams);

  it++;
  if(it > NumTimeSteps+1) {
   fprintf(stderr,"SetCurrentTimetoNext(): Already beyond horizon\n");
   exit(1);
  } else {
   PathParams->CurrentTimeIndex=it;
  }

}



int BeyondHorizon(PATH *PathParams )
{
  int it = GetCurrentTimeIndex(PathParams);
  int NumTimeSteps=HowManyTimeSteps(PathParams);

  if(it > NumTimeSteps) {
   return 1;
  } else {
   return 0;
  }

}



int OnHorizon( PATH *PathParams )
{
  int it = GetCurrentTimeIndex(PathParams);
  int NumTimeSteps=HowManyTimeSteps(PathParams);

  if(it == NumTimeSteps) {
   return 1;
  } else {
   return 0;
  }

}



void SetCurrentRatetoFirst( PATH *PathParams )
{
 PathParams->CurrentRate = 1;
}



void SetCurrentRateIndex(PATH *PathParams, int ir )
{
 PathParams->CurrentRate = ir;
}



int FinishedLastRate( PATH *PathParams )
{
  int NumRates=HowManyRates(PathParams);
  int ir = GetCurrentRateIndex(PathParams);

  if(ir > NumRates) {
   return 1;
  } else {
   return 0;
  }
}



int GetCurrentRateIndex( PATH *PathParams)
{
   return PathParams->CurrentRate;
}



void SetCurrentRatetoNext( PATH *PathParams)
{
  int NumRates=HowManyRates(PathParams);
  int ir = GetCurrentRateIndex(PathParams);

  ir++;
  if(ir > NumRates+1) {
   fprintf(stderr,"SetCurrentRatetoNext(): Already beyond last rate\n");
   exit(1);
  } else {
   PathParams->CurrentRate=ir;
  }

}



#endif



#ifndef HISTORY



PATH *CreatePath
               ( 
               int NumTimeSteps,
               int NumRates, 
               int Lag
               )
{
  int i;
  int err = 0;
  PATH *PathParams = NULL;

  if((PathParams=(PATH *)malloc(sizeof(PATH))) == NULL) {
   perror("CreatePath(): cannot allocate buffer for path\n");
   exit(1);
  }

  if((PathParams->History=(REAL **)
      malloc((NumTimeSteps+1)*sizeof(REAL *))) == NULL) {
      fprintf(stderr,"cannot allocate buffer for rate pointers in a path \n");
      exit(1);
  }

  for(i=0;i<=NumTimeSteps;i++) {
    if((PathParams->History[i]=(REAL *)
      malloc((NumRates+1)*sizeof(REAL))) == NULL) {
      fprintf(stderr,"cannot allocate buffer for time step # %d\n",i);
      exit(1);
    }
  }


  if((PathParams->ValidRate=(int **)
      malloc((NumTimeSteps+1)*sizeof(int *))) == NULL) {
      fprintf(stderr,"cannot allocate buffer for valid flag pointers in a path \n");
      exit(1);
  }

  for(i=0;i<=NumTimeSteps;i++) {
    if((PathParams->ValidRate[i]=(int *)
      malloc((NumRates+1)*sizeof(int))) == NULL) {
      fprintf(stderr,"cannot allocate buffer for flags for time step # %d\n",i);
      exit(1);
    }
  }

  if((PathParams->AllValidRates=(int *)
      malloc((NumTimeSteps+1)*sizeof(int))) == NULL) {
      fprintf(stderr,"cannot allocate buffer for all valid flags \n");
      exit(1);
  }

  if(Lag > 0) {

     if((PathParams->InitLags=(REAL **)
         malloc((NumRates+1)*sizeof(REAL *))) == NULL) {
         fprintf(stderr,"cannot allocate buffer for initial rates \n");
         exit(1);
     }
     for(i=0;i<=Lag;i++) {
       if((PathParams->InitLags[i]=(REAL *)
         malloc((NumRates+1)*sizeof(REAL))) == NULL) {
         fprintf(stderr,"cannot allocate buffer for initial lag # %d\n",i);
         exit(1);
       }
     }

  } else {
    PathParams->InitLags = NULL;
  }

  SetTimeGridSize(PathParams, NumTimeSteps );
  SetRateGridSize(PathParams,  NumRates );
  PathParams->Lag = Lag;


  return PathParams;
}



void PATH_Free(PATH *PathParams)
{
  int i;
  int  NumTimeSteps;
  int Lag; 

  if(PathParams == NULL) {
    fprintf(stderr,"PATH_Free(): cannot free a null path\n");
    exit(1);
  }

  NumTimeSteps = HowManyTimeSteps(PathParams);
  Lag = ProcessOrder(PathParams);

  if(PathParams->History == NULL) {
    fprintf(stderr,"PATH_Free(): cannot free a null history buffer\n");
    exit(1);
  } else {
    for(i=0;i<=NumTimeSteps;i++) {
      if(PathParams->History[i]== NULL) {
        fprintf(stderr,"PATH_Free(): cannot free a null buffer for time step# %d\n",i);
        exit(1);
      } else {
        free((void *)PathParams->History[i]);
      }
    }
    free((void *)PathParams->History);
  }

  if(Lag > 0) {
    if(PathParams->InitLags == NULL) {
      fprintf(stderr,"PATH_Free(): cannot free a null init lag buffer\n");
      exit(1);
    } else {
      for(i=0;i<=Lag;i++) {
        if(PathParams->InitLags[i]== NULL) {
          fprintf(stderr,"PATH_Free(): cannot free a null lag buffer # %d\n",i);
          exit(1);
        } else {
          free((void *)PathParams->InitLags[i]);
        }
      }
      free((void *)PathParams->InitLags);
    }
  }

  free((void *)PathParams);

}



int GetCurrentTimeSlice
                      (
                      PATH* PathParams, 
                      REAL *p 
                      )
{

  int ir;
  int  NumRates         = HowManyRates(PathParams);
  int NumTimeSteps     = HowManyTimeSteps(PathParams);
  int it = GetCurrentTimeIndex(PathParams);

  if(p != NULL) {
     for(ir=0; ir <= NumRates; ir ++)
        *(p+ir) = *(PathParams->History[it] + ir);
  } else {
     fprintf(stderr, "GetCurrentTimeSlice(): Invalid pointer argument\n");
     exit(1);
  }

  return DONE;

}



int GetCurrentMCStep
                   ( 
                   PATH *PathParams, 
                   REAL *p 
                   )
{

  int ir;
  int it = HowManyTimeSteps(PathParams);
  int NumRates  = HowManyRates(PathParams);

  if(p != NULL) {
     it = HowManyTimeSteps(PathParams);
     for(ir=0; ir <= NumRates; ir ++)
        *(p+ir) = *(PathParams->History[it] + ir);
  } else {
     fprintf(stderr, "GetCurrentMCStep(): Invalid pointer argument\n");
     exit(1);
  }

  return DONE;
}



int GetPreviousTimeSlice
                       ( 
                       PATH *PathParams, 
                       REAL *p 
                       )
{
  int ir;
  int NumRates         = HowManyRates(PathParams);
  int NumTimeSteps     = HowManyTimeSteps(PathParams);
  int it = GetPreviousTimeIndex(PathParams);


  if(p != NULL) {
  for(ir=0; ir <= NumRates; ir ++)
     *(p+ir) = *(PathParams->History[it] + ir);
  } else {
     fprintf(stderr, "GetPreviousTimeSlice(): Invalid pointer argument\n");
     exit(1);
  }

  return DONE;
}



int GetNextTimeSlice
                   ( 
                   PATH *PathParams, 
                   REAL *p 
                   )
{
  int ir;
  int NumRates         = HowManyRates(PathParams);
  int NumTimeSteps     = HowManyTimeSteps(PathParams);
  int it = GetNextTimeIndex(PathParams);


  if(p != NULL) {
  for(ir=0; ir <= NumRates; ir ++)
     *(p+ir) = *(PathParams->History[it] + ir);
  } else {
     fprintf(stderr, "GetPreviousTimeSlice(): Invalid pointer argument\n");
     exit(1);
  }

  return DONE;
}



void FlipAllHistories(PATH *PathParams)
{
  int ir;
  int it;
  int  NumRates=HowManyRates(PathParams);
  int NumTimeSteps= HowManyTimeSteps(PathParams);

  for(it=1;it<=NumTimeSteps;it++)
    for(ir=1; ir<=NumRates; ir++) 
      *(PathParams->History[it] + ir) *= (-1);

}



void UpdateCurrentTimeSlice
                          (
                          PATH *PathParams, 
                          REAL *NewRates
                          )
{
  int ir;
  int NumRates= HowManyRates(PathParams);
  int it=GetCurrentTimeIndex(PathParams);
 
  for (ir=1;ir<=NumRates;ir++)
              *(PathParams->History[it]+ir) = *(NewRates+ir);

}



void ShiftCurrentRateHistory
                           (
                           PATH *PathParams, 
                           REAL shift
                           )
{
  int it1;
  int NumTimeSteps= HowManyTimeSteps(PathParams);
  int it=GetCurrentTimeIndex(PathParams);
  int ir=GetCurrentRateIndex(PathParams);

  for (it1=it;it1<=NumTimeSteps;it1++)
              *(PathParams->History[it1]+ir) += shift;

}



void ShiftCurrentRate
                    (
                    PATH *PathParams, 
                    REAL shift
                    )
{
  int it1;
  int it=GetCurrentTimeIndex(PathParams);
  int ir=GetCurrentRateIndex(PathParams);

  *(PathParams->History[it]+ir) += shift;

}



int SetCurrentPath
                 (
                 PATH *PathParams, 
                 REAL *p
                 ) 
{
  int it;
  int ir=GetCurrentRateIndex(PathParams);
  int NumTimeSteps= HowManyTimeSteps(PathParams);

  if(PathParams == NULL) {
     fprintf(stderr,"SetCurrentPath(): PathParams not allocated\n");
     exit(1);
  }
  for(it=0;it<=NumTimeSteps;it++) *(PathParams->History[it]+ir)= *(p+it);

  return 0;

}



void GetCurrentPath
                  (
                  PATH *PathParams, 
                  REAL *p
                  )
{
  int it;
  int ir;
  int NumTimeSteps;

  if(PathParams == NULL) {
     fprintf(stderr,"GetCurrentPath(): PathParams not allocated\n");
     exit(1);
  }

  ir=GetCurrentRateIndex(PathParams);
  NumTimeSteps= HowManyTimeSteps(PathParams);

  for(it=0;it<=NumTimeSteps;it++) *(p+it) = *(PathParams->History[it]+ir);

}


void GetInitLagsofCurrentPath
                       (
                       PATH *PathParams, 
                       REAL *p
                       )
{
  int it;
  int ir=GetCurrentRateIndex(PathParams);
  int NumTimeSteps= HowManyTimeSteps(PathParams);

  if(PathParams == NULL) {
     fprintf(stderr,"GetCurrentPath(): PathParams not allocated\n");
     exit(1);
  }
  for(it=0;it<=NumTimeSteps;it++) *(p+it) = *(PathParams->InitLags[it]+ir);

}



void SetInitLagsofCurrentPath
                            (
                            PATH *PathParams, 
                            REAL *p
                            )
{
  int it;
  int ir=GetCurrentRateIndex(PathParams);
  int NumTimeSteps= HowManyTimeSteps(PathParams);

  if(PathParams == NULL) {
     fprintf(stderr,"GetCurrentPath(): PathParams not allocated\n");
     exit(1);
  }

  for(it=0;it<=NumTimeSteps;it++) *(PathParams->InitLags[it]+ir) = *(p+it);

}



int SetStraightPath
                  (
                  PATH *PathParams, 
                  REAL c
                  ) 
{
  int it;
  int ir=GetCurrentRateIndex(PathParams);
  int NumTimeSteps= HowManyTimeSteps(PathParams);

  if(PathParams == NULL) {
     fprintf(stderr,"SetCurrentPath(): PathParams not allocated\n");
     exit(1);
  }

  for(it=0;it<=NumTimeSteps;it++) *(PathParams->History[it]+ir)= c;

  return 0;
}



void SetAllPathsToZero(PATH *PathParams)
{
  int it;
  int ir;
  int NumTimeSteps= HowManyTimeSteps(PathParams);
  int NumRates = HowManyRates(PathParams);

  if(PathParams == NULL) {
     fprintf(stderr,"SetCurrentPath(): PathParams not allocated\n");
     exit(1);
  }
  for(it=0;it<=NumTimeSteps;it++) 
    for(ir=0;ir<=NumRates;ir++) 
      *(PathParams->History[it]+ir)= 0.0;

}



int SetCurrentTimeSlice
                      (
                      PATH *PathParams, 
                      REAL *p
                      ) 
{
  int ir;
  int it=GetCurrentTimeIndex(PathParams);
  int NumRates= HowManyRates(PathParams);

  if(PathParams == NULL) {
     fprintf(stderr,"SetCurrentTimeSlice(): PathParams not allocated\n");
     exit(1);
  }

  for(ir=0;ir<=NumRates;ir++) {
    *(PathParams->History[it]+ir)= *(p+ir);
  }

  return 0;
}



int SetInitLagsToZero(PATH *PathParams)
{
  int ir;
  int il;
  int NumRates;
  int Lag;
  int err = 0;

  if(PathParams == NULL) {
     fprintf(stderr,"SetInitRatesToZero(): PathParams not allocated\n");
     exit(1);
  }

  NumRates=HowManyRates(PathParams);
  Lag=PathParams->Lag;

  if(PathParams->InitLags == NULL) {
   fprintf(stderr,"SetInitRatesToZero(): PathParams->InitLags not allocated\n");
   exit(1);
  }

  for(il=1;il<=Lag;il++) {
    for(ir=1;ir<=NumRates;ir++) {
       *(PathParams->InitLags[il] + ir)=0.0;
     }
  }
  return 0;
}



int SetAllInitLags
                 (
                 PATH *PathParams, 
                 REAL **p
                 )
{
  int ir;
  int il;
  int NumRates;
  int Lag;
  int err = 0;	

  if(PathParams == NULL) {
     fprintf(stderr,"SetInitRates(): PathParams not allocated\n");
     exit(1);
  }
  NumRates=HowManyRates(PathParams);
  Lag=PathParams->Lag;

  if(PathParams->InitLags == NULL) {
   fprintf(stderr,"SetInitRates(): PathParams->InitRate not allocated\n");
   exit(1);
  }

  for(il=1;il<=Lag;il++) {
    for(ir=1;ir<=NumRates;ir++) {
      *(PathParams->InitLags[il] +ir)= *(p[ir]+il);
    }
  }

  return err;
}



int SetInitLag
             (
             int il, 
             PATH *PathParams, 
             REAL *p
             )
{
  int ir;
  int NumRates;
  int Lag;
  int err = 0;

  if(PathParams == NULL) {
     fprintf(stderr,"SetInitRates(): PathParams not allocated\n");
     exit(1);
  }

  NumRates=HowManyRates(PathParams);
  Lag=PathParams->Lag;

  if(PathParams->InitLags == NULL) {
   fprintf(stderr,"SetInitLag(): PathParams->InitLags not allocated\n");
   exit(1);
  }

  for(ir=1;ir<=NumRates;ir++) {
      *(PathParams->InitLags[il] +ir)= *(p+ir);
  }

  return err;
}



void SetTimeGridSize
                   (
                   PATH *PathParams, 
                   int NumTimeSteps 
                   )
{
  if(NumTimeSteps < 1) {
    fprintf(stderr,"SetTimeGridSize(): Unacceptable number of time steps ... %d\n", NumTimeSteps);
    exit(1);
  }

  PathParams->TimeGridSize = NumTimeSteps;
}



void SetRateGridSize
                   (
                   PATH *PathParams, 
                   int NumRates 
                   )
{
  if(NumRates < 1) {
    fprintf(stderr,"SetRateGridSize(): Unacceptable number of rates ... %d\n", NumRates);
    exit(1);
  }

  PathParams->RateGridSize = NumRates;
}



void FetchRunningTimeSlice
                         (
                         PATH *PathParams, 
                         REAL *p
                         )
{
 int NumRates = HowManyRates(PathParams);
 int itime = PathParams->RunningTimeIndex;
 int it;
 int ir;

 if(itime <= 0) {
   it = 1- itime;
   for(ir=1;ir<=NumRates;ir++) *(p+ir) = *(PathParams->InitLags[it] + ir);
 } else {
   for(ir=1;ir<=NumRates;ir++) *(p+ir) = *(PathParams->History[itime] + ir);
 }

}
   


void PATH_Add
            (
            PATH *Path1, 
            PATH *Path2
            )
{
  int err = 0;
  int ir;
  int  NumRates; 
  int size;
  REAL *Slice1;
  REAL *Slice2;

  if(Path1 == NULL || Path2 == NULL) {
     fprintf(stderr,"PATH_Add(): Invalid argument\n");
     exit(1);
  }

  if((NumRates=HowManyRates(Path1)) != HowManyRates(Path2)) {
     fprintf(stderr,"PATH_Add(): Incompatible paths\n");
     exit(1);
  }

  size = (NumRates + 1)*sizeof(REAL);
  Slice1 = (REAL *) mymalloc(size,"PATH_Add():Slice1\n");
  Slice2 = (REAL *) mymalloc(size,"PATH_Add():Slice2\n");

  SetCurrentTimetoInitial(Path1);
  SetCurrentTimetoInitial(Path2);
  SetCurrentTimetoNext(Path1);
  SetCurrentTimetoNext(Path2);

  while((!BeyondHorizon(Path1)) && (!BeyondHorizon(Path2))) {

       if( GetCurrentTimeSlice(Path1, Slice1) != DONE) {
         fprintf(stderr,"GetCurrentTimeSlice(): Time Slice # %d\n", GetCurrentTimeIndex(Path1));
         exit(1);
       }
       if( GetCurrentTimeSlice(Path2, Slice2) != DONE) {
         fprintf(stderr,"GetCurrentTimeSlice(): Time Slice # %d\n", GetCurrentTimeIndex(Path2));
         exit(1);
       }
       for(ir=1;ir<=NumRates;ir++) *(Slice1 + ir) += *(Slice2 + ir);

       UpdateCurrentTimeSlice(Path1,Slice1);
       SetCurrentTimetoNext(Path1);
       SetCurrentTimetoNext(Path2);
  }

}



#else



PATH *CreatePath
               ( 
               int NumTimeSteps,
               int NumRates,
               int  Lag
               )
{
  int i;
  int err = 0;
  PATH *PathParams = NULL;

  if((PathParams=(PATH *)malloc(sizeof(PATH))) == NULL) {
   perror("CreatePath(): cannot allocate buffer for path\n");
   exit(1);
  }

  if((PathParams->History=(REAL **)
      malloc((NumRates+1)*sizeof(REAL *))) == NULL) {
      fprintf(stderr,"cannot allocate buffer for rate pointers in a path \n");
      exit(1);
  }

  for(i=0;i<=NumRates;i++) {
    if((PathParams->History[i]=(REAL *)
      malloc((NumTimeSteps+1)*sizeof(REAL))) == NULL) {
      fprintf(stderr,"cannot allocate buffer for rate # %d\n",i);
      exit(1);
    }
  }

  if((PathParams->ValidRate=(int **)
      malloc((NumRates+1)*sizeof(int *))) == NULL) {
      fprintf(stderr,"cannot allocate buffer for valid flag pointers in a path \n");
      exit(1);
  }

  for(i=0;i<=NumRates;i++) {
    if((PathParams->ValidRate[i]=(int *)
      malloc((NumTimeSteps+1)*sizeof(int))) == NULL) {
      fprintf(stderr,"cannot allocate buffer for flags for rate # %d\n",i);
      exit(1);
    }
  }

  if((PathParams->AllValidRates=(int *)
      malloc((NumTimeSteps+1)*sizeof(int))) == NULL) {
      fprintf(stderr,"cannot allocate buffer for all valid flags \n");
      exit(1);
  }

  if(Lag > 0) {

     if((PathParams->InitLags=(REAL **)
         malloc((NumRates+1)*sizeof(REAL *))) == NULL) {
         fprintf(stderr,"cannot allocate buffer for initial rates \n");
         exit(1);
     }

     for(i=0;i<=NumRates;i++) {
       if((PathParams->InitLags[i]=(REAL *)
         malloc((Lag+1)*sizeof(REAL))) == NULL) {
         fprintf(stderr,"cannot allocate buffer for initial lag # %d\n",i);
         exit(1);
       }
     }

  } else {

    PathParams->InitLags = NULL;

  }

  SetTimeGridSize(PathParams, NumTimeSteps );
  SetRateGridSize(PathParams,  NumRates );
  PathParams->Lag = Lag;


  return PathParams;
}



void PATH_Free(PATH *PathParams)
{
  int i;
  int  NumRates;
  int Lag; 

  if(PathParams == NULL) {
    fprintf(stderr,"PATH_Free(): cannot free a null path\n");
    exit(1);
  }

  NumRates = HowManyRates(PathParams);
  Lag = ProcessOrder(PathParams);

  if(PathParams->History == NULL) {
    fprintf(stderr,"PATH_Free(): cannot free a null history buffer\n");
    exit(1);
  } else {
    for(i=0;i<=NumRates;i++) {
      if(PathParams->History[i]== NULL) {
        fprintf(stderr,"PATH_Free(): cannot free a null buffer for rate # %d\n",i);
        exit(1);
      } else {
        free((void *)PathParams->History[i]);
      }
    }
    free((void *)PathParams->History);
  }

  if(Lag > 0) {
    if(PathParams->InitLags == NULL) {
      fprintf(stderr,"PATH_Free(): cannot free a null init lag buffer\n");
      exit(1);
    } else {
      for(i=0;i<=NumRates;i++) {
        if(PathParams->InitLags[i]== NULL) {
          fprintf(stderr,"PATH_Free(): cannot free a null lag buffer for rate # %d\n",i);
          exit(1);
        } else {
          free((void *)PathParams->InitLags[i]);
        }
      }
      free((void *)PathParams->InitLags);
    }
  }

  free((void *)PathParams);
}



int GetCurrentTimeSlice
                      (
                      PATH* PathParams, 
                      REAL *p 
                      )
{
  int ir;
  int NumRates         = HowManyRates(PathParams);
  int NumTimeSteps     = HowManyTimeSteps(PathParams);
  int it = GetCurrentTimeIndex(PathParams);

  if(p != NULL) {
     for(ir=0; ir <= NumRates; ir ++)
        *(p+ir) = *(PathParams->History[ir] + it);
  } else {
     fprintf(stderr, "GetCurrentTimeSlice(): Invalid pointer argument\n");
     exit(1);
  }

  return DONE;

}



int GetCurrentMCStep
                   ( 
                   PATH *PathParams, 
                   REAL *p 
                   )
{

  int ir;
  int NumRates = HowManyRates(PathParams);
  int it = HowManyTimeSteps(PathParams);

  if(p != NULL) {
     for(ir=0; ir <= NumRates; ir ++)
        *(p+ir) = *(PathParams->History[ir] + it);
  } else {
     fprintf(stderr, "GetCurrentMCStep(): Invalid pointer argument\n");
     exit(1);
  }

  return DONE;
}



int GetPreviousTimeSlice
                       ( 
                       PATH *PathParams, 
                       REAL *p 
                       )
{
  int ir;
  int NumRates         = HowManyRates(PathParams);
  int NumTimeSteps     = HowManyTimeSteps(PathParams);
  int it = GetPreviousTimeIndex(PathParams);


  if(p != NULL) {
  for(ir=0; ir <= NumRates; ir ++)
     *(p+ir) = *(PathParams->History[ir] + it);
  } else {
     fprintf(stderr, "GetPreviousTimeSlice(): Invalid pointer argument\n");
     exit(1);
  }

  return DONE;
}



int GetNextTimeSlice
                   ( 
                   PATH *PathParams, 
                   REAL *p 
                   )
{

  int ir;
  int NumRates         = HowManyRates(PathParams);
  int NumTimeSteps     = HowManyTimeSteps(PathParams);
  int it = GetNextTimeIndex(PathParams);


  if(p != NULL) {
  for(ir=0; ir <= NumRates; ir ++)
     *(p+ir) = *(PathParams->History[ir] + it);
  } else {
     fprintf(stderr, "GetPreviousTimeSlice(): Invalid pointer argument\n");
     exit(1);
  }

  return DONE;
}



void FlipAllHistories(PATH *PathParams)
{
  int ir;
  int it;
  int  NumRates=HowManyRates(PathParams);
  int NumTimeSteps= HowManyTimeSteps(PathParams);

  for(ir=1; ir<=NumRates; ir++) for(it=1;it<=NumTimeSteps;it++)
    *(PathParams->History[ir] + it) *= (-1);
}



void UpdateCurrentTimeSlice
                          (
                          PATH *PathParams, 
                          REAL *NewRates
                          )
{
  int ir;
  int NumRates= HowManyRates(PathParams);
  int it=GetCurrentTimeIndex(PathParams);
 
  for (ir=1;ir<=NumRates;ir++)
              *(PathParams->History[ir]+it) = *(NewRates+ir);

}



void ShiftCurrentRateHistory
                           (
                           PATH *PathParams, 
                           REAL shift
                           )
{
  int it1;
  int NumTimeSteps= HowManyTimeSteps(PathParams);
  int it=GetCurrentTimeIndex(PathParams);
  int ir=GetCurrentRateIndex(PathParams);

  for (it1=it;it1<=NumTimeSteps;it1++)
              *(PathParams->History[ir]+it1) += shift;

}



void ShiftCurrentRate
                    (
                    PATH *PathParams, 
                    REAL shift
                    )
{
  int it1;
  int it=GetCurrentTimeIndex(PathParams);
  int ir=GetCurrentRateIndex(PathParams);

  *(PathParams->History[ir]+it) += shift;

}



int SetCurrentPath
                 (
                 PATH *PathParams, 
                 REAL *p
                 ) 
{
  int it;
  int ir;
  int NumTimeSteps;

  if(PathParams == NULL) {
     fprintf(stderr,"SetCurrentPath(): PathParams not allocated\n");
     exit(1);
  }

  NumTimeSteps = HowManyTimeSteps(PathParams);
  ir = GetCurrentRateIndex(PathParams);

  if(PathParams->History[ir] == NULL) {
     fprintf(stderr,"SetCurrentPath(): PathParams->History[%d] not allocated\n", ir);
     exit(1);
  }
  for(it=0;it<=NumTimeSteps;it++) *(PathParams->History[ir]+it)= *(p+it);

  return 0;
}



void GetCurrentPath
                  (
                  PATH *PathParams, 
                  REAL *p
                  )
{
  int it;
  int ir=GetCurrentRateIndex(PathParams);
  int NumTimeSteps= HowManyTimeSteps(PathParams);

  if(PathParams == NULL) {
     fprintf(stderr,"GetCurrentPath(): PathParams not allocated\n");
     exit(1);
  }

  if(PathParams->History[ir] == NULL) {
     fprintf(stderr,"GetCurrentPath(): PathParams->History[%d] not allocated\n", ir);
     exit(1);
  }

  for(it=0;it<=NumTimeSteps;it++) *(p+it) = *(PathParams->History[ir]+it);

}



void GetInitLagsofCurrentPath
                            (
                            PATH *PathParams, 
                            REAL *p
                            )
{
  int it;
  int ir=GetCurrentRateIndex(PathParams);
  int NumTimeSteps= HowManyTimeSteps(PathParams);

  if(PathParams == NULL) {
     fprintf(stderr,"GetCurrentPath(): PathParams not allocated\n");
     exit(1);
  }

  if(PathParams->InitLags[ir] == NULL) {
     fprintf(stderr,"GetCurrentPath(): PathParams->Lags[%d] not allocated\n", ir);
     exit(1);
  }

  for(it=0;it<=NumTimeSteps;it++) *(p+it) = *(PathParams->InitLags[ir]+it);

}



void SetInitLagsofCurrentPath
                            (
                            PATH *PathParams, 
                            REAL *p
                            )
{
  int it;
  int ir=GetCurrentRateIndex(PathParams);
  int NumTimeSteps= HowManyTimeSteps(PathParams);

  if(PathParams == NULL) {
     fprintf(stderr,"GetCurrentPath(): PathParams not allocated\n");
     exit(1);
  }

  if(PathParams->InitLags[ir] == NULL) {
     fprintf(stderr,"GetCurrentPath(): PathParams->Lags[%d] not allocated\n", ir);
     exit(1);
  }

  for(it=0;it<=NumTimeSteps;it++) *(PathParams->InitLags[ir]+it) = *(p+it);

}



int SetStraightPath
                  (
                  PATH *PathParams, 
                  REAL c
                  ) 
{
  int it;
  int ir=GetCurrentRateIndex(PathParams);
  int NumTimeSteps= HowManyTimeSteps(PathParams);

  if(PathParams == NULL) {
     fprintf(stderr,"SetCurrentPath(): PathParams not allocated\n");
     exit(1);
  }

  if(PathParams->History[ir] == NULL) {
     fprintf(stderr,"SetCurrentPath(): PathParams->History[%d] not allocated\n", ir);
     exit(1);
  }

  for(it=0;it<=NumTimeSteps;it++) *(PathParams->History[ir]+it)= c;

  return 0;
}



void SetAllPathsToZero(PATH *PathParams)
{
  int it;
  int ir;
  int NumTimeSteps= HowManyTimeSteps(PathParams);
  int NumRates = HowManyRates(PathParams);

  if(PathParams == NULL) {
     fprintf(stderr,"SetCurrentPath(): PathParams not allocated\n");
     exit(1);
  }
  for(ir=0;ir<=NumRates;ir++) {
    if(PathParams->History[ir] == NULL) {
       fprintf(stderr,"SetCurrentPath(): History[%d] not allocated\n", ir);
       exit(1);
    }
    for(it=0;it<=NumTimeSteps;it++) *(PathParams->History[ir]+it)= 0.0;
  }

}



int SetCurrentTimeSlice
                      (
                      PATH *PathParams, 
                      REAL *p
                      ) 
{
  int ir;
  int it=GetCurrentTimeIndex(PathParams);
  int NumRates= HowManyRates(PathParams);

  if(PathParams == NULL) {
     fprintf(stderr,"SetCurrentTimeSlice(): PathParams not allocated\n");
     exit(1);
  }

  for(ir=0;ir<=NumRates;ir++) {
    if(PathParams->History[ir] == NULL) {
       fprintf(stderr,"SetCurrentTimeSlice(): History[%d] not allocated\n", ir);
       exit(1);
    }
    *(PathParams->History[ir]+it)= *(p+ir);
  }

  return 0;
}



int SetInitLagsToZero(PATH *PathParams)
{
  int ir;
  int il;
  int NumRates;
  int Lag;
  int err = 0;

  if(PathParams == NULL) {
     fprintf(stderr,"SetInitRatesToZero(): PathParams not allocated\n");
     exit(1);
  }
  NumRates=HowManyRates(PathParams);
  Lag=PathParams->Lag;

  if(PathParams->InitLags == NULL) {
   fprintf(stderr,"SetInitRatesToZero(): PathParams->InitLags not allocated\n");
   exit(1);
  }

  for(ir=1;ir<=NumRates;ir++) {
     for(il=1;il<=Lag;il++) {
       *(PathParams->InitLags[ir] + il)=0.0;
     }
  }
  return 0;
}



int SetAllInitLags
                 (
                 PATH *PathParams, 
                 REAL **p
                 )
{
  int ir;
  int il;
  int NumRates;
  int Lag;
  int err = 0;	

  if(PathParams == NULL) {
     fprintf(stderr,"SetInitRates(): PathParams not allocated\n");
     exit(1);
  }

  NumRates=HowManyRates(PathParams);
  Lag=PathParams->Lag;

  if(PathParams->InitLags == NULL) {
   fprintf(stderr,"SetInitRates(): PathParams->InitRate not allocated\n");
   exit(1);
  }

  for(ir=1;ir<=NumRates;ir++) {
   for(il=1;il<=Lag;il++) {
      *(PathParams->InitLags[ir] +il)= *(p[ir]+il);
   }
  }

  return err;
}



int SetInitLag
             (
             int il, 
             PATH *PathParams, 
             REAL *p
             )
{
  int ir;
  int NumRates;
  int Lag;
  int err = 0;

  if(PathParams == NULL) {
     fprintf(stderr,"SetInitRates(): PathParams not allocated\n");
     exit(1);
  }

  NumRates=HowManyRates(PathParams);
  Lag=PathParams->Lag;

  if(PathParams->InitLags == NULL) {
   fprintf(stderr,"SetInitLag(): PathParams->InitLags not allocated\n");
   exit(1);
  }

  for(ir=1;ir<=NumRates;ir++) {
      *(PathParams->InitLags[ir] +il)= *(p+ir);
  }

  return err;
}


void SetTimeGridSize
                   (
                   PATH *PathParams, 
                   int NumTimeSteps 
                   )
{
  if(NumTimeSteps < 1) {
    fprintf(stderr,"SetTimeGridSize(): Unacceptable number of time steps ... %d\n", NumTimeSteps);
    exit(1);
  }

  PathParams->TimeGridSize = NumTimeSteps;
}



void SetRateGridSize
                   (
                   PATH *PathParams, 
                   int NumRates 
                   )
{
  if(NumRates < 1) {
    fprintf(stderr,"SetRateGridSize(): Unacceptable number of rates ... %d\n", NumRates);
    exit(1);
  }

  PathParams->RateGridSize = NumRates;
}



void FetchRunningTimeSlice
                         (
                         PATH *PathParams, 
                         REAL *p
                         )
{
 int NumRates = HowManyRates(PathParams);
 int itime = PathParams->RunningTimeIndex;
 int it;
 int ir;

 if(itime <= 0) {
   it = 1- itime;
   for(ir=1;ir<=NumRates;ir++) *(p+ir) = *(PathParams->InitLags[ir] + it);
 } else {
   for(ir=1;ir<=NumRates;ir++) *(p+ir) = *(PathParams->History[ir] + itime);
 }

}
   


void PATH_Add
            (
            PATH *Path1, 
            PATH *Path2
            )
{
  int err = 0;
  int ir;
  int  NumRates;
  int  size;
  REAL *Slice1;
  REAL *Slice2;

  if(Path1 == NULL || Path2 == NULL) {
     fprintf(stderr,"PATH_Add(): Invalid argument\n");
     exit(1);
  }

  if((NumRates=HowManyRates(Path1)) != HowManyRates(Path2)) {
     fprintf(stderr,"PATH_Add(): Incompatible paths\n");
     exit(1);
  }

  size = (NumRates + 1)*sizeof(REAL);
  Slice1 = (REAL *) mymalloc(size,"PATH_Add():Slice1\n");
  Slice2 = (REAL *) mymalloc(size,"PATH_Add():Slice2\n");

  SetCurrentTimetoInitial(Path1);
  SetCurrentTimetoInitial(Path2);
  SetCurrentTimetoNext(Path1);
  SetCurrentTimetoNext(Path2);

  while((!BeyondHorizon(Path1)) && (!BeyondHorizon(Path2))) {

       if( GetCurrentTimeSlice(Path1, Slice1) != DONE) {
         fprintf(stderr,"GetCurrentTimeSlice(): Time Slice # %d\n", GetCurrentTimeIndex(Path1));
         exit(1);
       }

       if( GetCurrentTimeSlice(Path2, Slice2) != DONE) {
         fprintf(stderr,"GetCurrentTimeSlice(): Time Slice # %d\n", GetCurrentTimeIndex(Path2));
         exit(1);
       }

       for(ir=1;ir<=NumRates;ir++) *(Slice1 + ir) += *(Slice2 + ir);

       UpdateCurrentTimeSlice(Path1,Slice1);
       SetCurrentTimetoNext(Path1);
       SetCurrentTimetoNext(Path2);

  }

}


#endif
